/*
 * RT-Linux POSIX condition variables implementation
 *
 * Written by Michael Barabanov
 * Copyright (C) VJY Associates LLC, 1999
 * Released under the terms of the GPL
 *
 */

#ifndef __RTL_CONDVAR__
#define __RTL_CONDVAR__

#ifdef __KERNEL__

#include <rtl_conf.h>
#ifdef CONFIG_RTL_MUTEX_SUPPORT

#include <rtl_mutex.h>

typedef struct {
	int pshared;
} pthread_condattr_t;

extern inline int pthread_condattr_getpshared(const pthread_condattr_t *attr,
		    int *pshared)
{
	*pshared = attr->pshared;
	return 0;
}

extern inline int pthread_condattr_setpshared(pthread_condattr_t *attr,
		    int pshared) {
	if (pshared != PTHREAD_PROCESS_SHARED
			&& pshared != PTHREAD_PROCESS_PRIVATE) {
		return EINVAL;
	}
	attr->pshared = pshared;
	return 0;
}


extern inline int pthread_condattr_init(pthread_condattr_t *attr)
{
	attr->pshared = PTHREAD_PROCESS_PRIVATE;
	return 0;
}

extern inline int pthread_condattr_destroy(pthread_condattr_t *attr)
{
	return 0;
}


typedef struct {
	spinlock_t lock;
	RTL_WQ wq;
} pthread_cond_t;



#define PTHREAD_COND_INITIALIZER { SPIN_LOCK_UNLOCKED, RTL_WQ_INITIALIZER }

extern int pthread_cond_init(pthread_cond_t *cond,
		    const pthread_condattr_t *attr);
extern int pthread_cond_destroy(pthread_cond_t *cond);

extern int pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
extern int pthread_cond_timedwait(pthread_cond_t *cond,
		    pthread_mutex_t *mutex, const struct timespec *abstime);

extern int pthread_cond_signal(pthread_cond_t *cond);
extern int pthread_cond_broadcast(pthread_cond_t *cond);

#endif
#endif


#endif
