/*
 * RT-Linux scheduler
 *
 * Written by Michael Barabanov, Victor Yodaiken
 * Copyright (C) VJY Associates LLC, 1998-1999
 * Released under the terms of the GPL
 *
 */

#ifndef __RTL__SCHED__
#define __RTL__SCHED__

#include <asm/ptrace.h>
#include <asm/spinlock.h>
#include <linux/malloc.h>
#include <linux/tqueue.h>

#include <rtl_conf.h>
#include <rtl_core.h>
#include <rtl_time.h>
#include <arch/rtl_fpu.h>
/* #include "../schedulers/x86/switch.h" */

extern void rtl_tq_sync (struct tq_struct *tq_s);

enum rtl_task_states {
	RTL_THREAD_READY = 1, RTL_THREAD_DELAYED = 2, RTL_THREAD_DORMANT = 4, RTL_THREAD_SUSPENDED = 8, RTL_THREAD_ACTIVE = 16, RTL_THREAD_ZOMBIE = 32, RTL_THREAD_WAIT_NEEDED = 64
};

/* for sched_param and others */
#include <linux/sched.h>

typedef struct sched_param rtl_sched_param;

#define TIMER_ABSTIME 1
#define RTL_MAX_TIMERS 32



typedef struct rtl_timer_struct *timer_t;

struct rtl_cleanup_struct {
	void (*routine)(void*);
	void *arg;
	struct rtl_cleanup_struct *next;
};

/* threads have no clocks. Clocks belong to schedulers.  We can add scheduling
   policies in which a scheduler juggles multiple clocks, but there is no advantage
   that I can see in allowing a thread to specify its  hardware clock
   */
#define RTL_THREAD_MAGIC 0x79433743
   
struct rtl_thread_struct {
	int *stack;	/* hardcoded */
	int uses_fp;
	enum rtl_task_states state;
	int *stack_bottom;
	rtl_sched_param sched_param;
	struct rtl_thread_struct *next;
	RTL_FPU_CONTEXT fpu_regs;
	int cpu;
	hrtime_t resume_time;
	hrtime_t period;
	void *retval;
	struct module *creator;
	int pending_signals;
	struct tq_struct free_task;
	void *user[4];
	int errno_val;
	struct rtl_cleanup_struct *cleanup;
	int magic;
	int pad [64];
};

#define RTL_MAX_SIGNAL 31 /* this is max for internal RTL signals */
/* these are bit positions */
#define RTL_SIGNAL_NULL 0 /* posix wants signal=0 to simply check */
#define RTL_SIGNAL_DELETE 1
#define RTL_SIGNAL_WAKEUP 2
#define RTL_SIGNAL_SUSPEND 3

/*TODO How will this work on PPC */
#define RTL_LINUX_MIN_SIGNAL 256  /* signals to Linux start here. global then local */
#define RTL_LINUX_MAX_SIGNAL 1024
#define RTL_LINUX_MIN_LOCAL_SIGNAL  512

extern void rtl_schedule (void);

typedef struct rtl_thread_struct RTL_THREAD_STRUCT;

typedef RTL_THREAD_STRUCT *pthread_t;
 

struct rtl_sched_cpu_struct {
	RTL_THREAD_STRUCT *rtl_current;
	RTL_THREAD_STRUCT rtl_linux_task;
	RTL_THREAD_STRUCT idle_task;
	RTL_THREAD_STRUCT *rtl_task_fpu_owner;	/* the task whose FP context is currently in the FPU unit */
	RTL_THREAD_STRUCT *rtl_tasks; /* the queue of  RT tasks */
	task_queue rtl_tq_cpu;
	clockid_t clock;
	hrtime_t timerset;
	spinlock_t rtl_tasks_lock;
	int sched_user[4];	/* on x86 sched_user[0] is the Linux TS flag */
}/* __attribute__ ((aligned (64)))*/;

typedef struct rtl_sched_cpu_struct  schedule_t;

#ifdef __SMP__
extern struct rtl_sched_cpu_struct rtl_sched [NR_CPUS];
#define sched_data(cpu) (&rtl_sched [cpu])
#define LOCAL_SCHED (&rtl_sched[rtl_getcpuid()])
#else
extern struct rtl_sched_cpu_struct rtl_sched [1];
#define sched_data(cpu) (&rtl_sched[0])
#define LOCAL_SCHED (&rtl_sched[0])
#endif

#define RTL_CURRENT (LOCAL_SCHED->rtl_current)

#define RTL_TQ(cpu) (sched_data(cpu)->rtl_tq_cpu)


/* RTL-specific function TODO: write POSIX equivalents for these */


extern int pthread_delete_np (pthread_t thread);

extern int pthread_setfp_np (pthread_t thread, int flag);

extern int pthread_wakeup_np (pthread_t thread);

extern int pthread_suspend_np (pthread_t thread);

/* end RTL-specific */


/* POSIX interface */

#define pthread_cleanup_push(p_routine, p_arg) \
{ \
	rtl_irqstate_t __flags; \
	struct rtl_cleanup_struct __cleanup; \
	__cleanup.routine = (p_routine); \
	__cleanup.arg = (p_arg); \
	rtl_no_interrupts (__flags); \
	__cleanup.next = pthread_self()->cleanup; \
	pthread_self()->cleanup = &__cleanup; \
	rtl_restore_interrupts (__flags);


#define pthread_cleanup_pop(execute) \
	rtl_no_interrupts (__flags); \
	pthread_self()->cleanup = pthread_self()->cleanup->next; \
	if (execute) \
		__cleanup.routine(__cleanup.arg); \
	rtl_restore_interrupts (__flags); \
}


extern inline int sched_get_priority_max(int policy) { return 1000000; }
extern inline int sched_get_priority_min(int policy) { return 0; }

extern inline pthread_t pthread_self(void) {
	return (LOCAL_SCHED)-> rtl_current;
}

extern inline int pthread_equal(pthread_t thread1, pthread_t thread2)
{
	return thread1 == thread2;
}

typedef struct STRUCT_PTHREAD_ATTR {
	int stack_size;
	rtl_sched_param sched_param;
	int cpu;
	int use_fp;
	 enum rtl_task_states initial_state;
} pthread_attr_t;

extern inline int pthread_attr_init(pthread_attr_t *attr)
{
	attr->stack_size = 8000;
	attr->sched_param.sched_priority = sched_get_priority_min(0);
	attr->cpu = rtl_getcpuid();
        attr->initial_state = RTL_THREAD_READY;
	attr->use_fp = 0;
	return 0;
}

extern inline int pthread_attr_destroy(pthread_attr_t *attr)
{
	return 0;
}

/*
extern inline int pthread_attr_setfp_np (pthread_attr_t *attr, int flag)
{
	attr->use_fp = flag;
	return 0;
}
*/

extern inline int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize)
{
	attr->stack_size = stacksize;
	return 0;
}

extern inline int pthread_attr_getcpu_np(pthread_attr_t *attr, int * cpu)
{
	*cpu = attr->cpu;
	return 0;
}

extern int pthread_attr_setcpu_np(pthread_attr_t *attr, int cpu);

extern int pthread_wait_np(void);

extern int pthread_make_periodic_np (pthread_t p, hrtime_t start_time, hrtime_t period);

/* this one is deprecated */
extern int pthread_setperiod_np(pthread_t p,const struct itimerspec *t);

extern inline clockid_t rtl_getschedclock(void)
{
	return LOCAL_SCHED->clock;
}

extern int rtl_setclockmode (clockid_t clock, int mode, hrtime_t period);


extern inline void pthread_yield(void)
{
/* TODO */
}


extern inline int pthread_setschedparam(pthread_t thread, int policy,
		    const rtl_sched_param *param) {
	thread->sched_param = *param;
	return 0;
}
 
extern inline int pthread_getschedparam(pthread_t thread, int *policy,
		    rtl_sched_param *param) {
	*param = thread->sched_param;
	return 0;
}

extern inline int pthread_attr_setschedparam(pthread_attr_t *attr,
		    const rtl_sched_param *param) {
	attr->sched_param = *param;
	return 0;
}

extern inline int pthread_attr_getschedparam(const pthread_attr_t *attr,
		    rtl_sched_param *param) {
	*param = attr->sched_param;
	return 0;
}
 


struct module;
extern struct module __this_module;
extern int __pthread_create (pthread_t *thread, pthread_attr_t *attr, void *(*start_routine)(void *), void *arg, struct module *mod);

#define pthread_create(thread, attr, start, arg) __pthread_create(thread, attr, start, arg, &__this_module)


#define PTHREAD_CANCELED ((void *) -1)

extern void pthread_exit(void *retval);

int pthread_kill(pthread_t , int signo);

/* end of POSIX interface */


#include <rtl_compat.h>

extern int rtl_sched_lock (pthread_t thread);

/* private interfaces */

/* #define CONFIG_RTL_FAST_RESCHED */
#ifdef CONFIG_RTL_FAST_RESCHED
extern void rtl_pend_reschedule(int cpu);
extern void rtl_do_reschedules(void);
#else
#define rtl_pend_reschedule(cpu)
#define rtl_do_reschedules()
#endif

#endif
