/*
 * RT-Linux scheduler compatibility functions
 *
 * Written by Michael Barabanov
 * simplification, Yodaiken  Dec 25 1999
 * Copyright (C) VJY Associates LLC, 1998,1999
 * Released under the terms of the GNU  General Public License
 *
 */

#include <linux/version.h>
#include <rtl_conf.h>
#ifdef CONFIG_RTL_USE_V1_API

#include <linux/errno.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/segment.h>

#include <rtl_time.h>
#include <rtl_sync.h>
#include <rtl_sched.h>
#include <rtl_tqueue.h>

int rt_task_init (RT_TASK *t, void (*fn)(int data), int data, int stack_size, int priority)
{
	int ret;
	void *(*f)(void *);
	pthread_attr_t attr;
	rtl_sched_param param;
	pthread_attr_init (&attr);
	f = (void *)fn;
	attr.initial_state = RTL_THREAD_DORMANT;
	attr.cpu = cpu_logical_map (0); /* all compatibility tasks execute on one CPU */
	pthread_attr_setstacksize(&attr, stack_size);
	param.sched_priority = sched_get_priority_max(0) - priority; /* note the priority scheme is inverted to correspond to the POSIX one */
	pthread_attr_setschedparam(&attr, &param);
	ret = pthread_create(t, &attr, f, (void *)data);
	return -ret;
}


int rt_task_suspend (RT_TASK *task)
{
	int v;
	v =  pthread_suspend_np(*task);
	return -v;
}


int rt_task_wakeup (RT_TASK *task)
{
	int v;
	v =  pthread_wakeup_np(*task);
        return -v;
}


int rt_task_make_periodic (RT_TASK *task, RTIME start_time, RTIME period)
{
	int v;
	rtl_irqstate_t flags;
	/* CONFIG_V1 should make this work like this, but
	   no CONFIG_V1 should allow this with nanoseconds */
	rtl_no_interrupts(flags);
	v = pthread_make_periodic_np (*task,
			start_time/**838*/, period/**838*/);
	(*task)->state = RTL_THREAD_DELAYED;
	rtl_schedule();
	rtl_restore_interrupts(flags);
	return -v;
}

#endif /* CONFIG_RTL_USE_V1_API */
