/*
 * RT-Linux mutex implementation
 *
 * Written by Michael Barabanov
 * Rewritten by Victor Yodaiken after a long delay.
 * Copyright (C) VJY Associates LLC, 1999,2000
 * Released under the terms of the GPL
 *
 */

#include <rtl_mutex.h>
#include <rtl_printf.h>
#ifdef CONFIG_RTL_MUTEX_SUPPORT

#include <rtl_sync.h>

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
      mutex->valid = 0;
      return 0;
}

int pthread_mutex_init(pthread_mutex_t *mutex,
		    const pthread_mutexattr_t *attr)
{
	int i;
	spin_lock_init (&mutex->lock);
        mutex->valid = 1;
	mutex->count = 0;
        for(i=0; i < CONFIG_RTL_MUTEX_MAX_WAITERS;i++)
		mutex->waiters[i]=0;
	if (attr) {
		pthread_mutexattr_gettype(attr, &mutex->type);
	} else {
		mutex->type = PTHREAD_MUTEX_DEFAULT;
	}
	return 0;
}


/* TODO: implement various mutex types. (No way! VY) */
int __pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	if (test_and_set_bit(0, &mutex->busy)) {
		return EBUSY;
	} else {
		mutex->owner = pthread_self();
		return 0;
	}
}


int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
/* 	rtl_irqstate_t flags; */
	int ret;
/* 	rtl_spin_lock_irqsave (&mutex->lock, flags); */

	ret = __pthread_mutex_trylock(mutex);

/* 	rtl_spin_unlock_irqrestore (&mutex->lock, flags); */
	return ret;
}

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
	pthread_t me = pthread_self();
	unsigned int flags;

	if (!mutex->valid) {
		return EINVAL;
	}
	rtl_spin_lock_irqsave (&mutex->lock, flags);
	while (__pthread_mutex_trylock(mutex)) {
		/* put ourselves on the queue */
		if(mutex->count >= CONFIG_RTL_MUTEX_MAX_WAITERS) {
			rtl_spin_unlock_irqrestore (&mutex->lock, flags);
			rtl_printf ("RTLinux mutex too many waiters: %d\n", mutex->count);
			return EBUSY; /*too many waiters ! */
		} else {
			mutex->waiters [mutex->count++] = me;
		}
		pthread_kill (me, RTL_SIGNAL_SUSPEND);

		rtl_spin_unlock_irqrestore(&mutex->lock, flags);
		rtl_schedule();
		rtl_spin_lock_irqsave(&mutex->lock, flags);

	}
	rtl_spin_unlock_irqrestore(&mutex->lock, flags);
	return 0;
}


int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	int i;
	rtl_irqstate_t flags;

	if (!mutex->valid) {
		return EINVAL;
	}

	rtl_spin_lock_irqsave (&mutex->lock, flags);
	for (i = 0; i < mutex->count; i++) {
		pthread_kill (mutex->waiters[i], RTL_SIGNAL_WAKEUP);
	}
	mutex->count = 0;
	clear_bit (0, &mutex->busy);
	rtl_spin_unlock_irqrestore (&mutex->lock, flags);
	return 0;
}


#endif
